import pprint
from dataclasses import dataclass
from typing import Optional

import numpy as np

from . import SurfacePoints, Orientations
from gempy_engine.core.data.generic_grid import GenericGrid
from gempy_engine.core.data.centered_grid import CenteredGrid
from .engine_grid import EngineGrid
from .regular_grid import RegularGrid
from .stack_relation_type import StackRelationType
from .stacks_structure import StacksStructure
from .kernel_classes.faults import FaultsData
from .kernel_classes.server.input_parser import InterpolationInputSchema


@dataclass
class InterpolationInput:
    # @ off
    surface_points: SurfacePoints
    orientations: Orientations
    _original_grid: EngineGrid
    
    @property
    def original_grid(self):
        return self._original_grid
    
    def set_grid_to_original(self):
        self._grid = self._original_grid


    _grid: EngineGrid
    @property
    def grid(self):
        return self._grid
    
    def set_temp_grid(self, value):
        self._grid = value

    _unit_values: Optional[np.ndarray] = None
    segmentation_function: Optional[callable] = None  # * From scalar field to values

    _all_surface_points: SurfacePoints = None

    # region per model ? Not sure what I mean here

    _fault_values: FaultsData = None  # ! This field is internal. If you want to pass fault data use StacksStructure
    stack_relation: StackRelationType = StackRelationType.ERODE  # ? Should be here or in the descriptor

    # endregion

    def __init__(self, surface_points: SurfacePoints, orientations: Orientations, grid: EngineGrid,
                 unit_values: Optional[np.ndarray] = None, segmentation_function: Optional[callable] = None,
                 stack_relation: StackRelationType = StackRelationType.ERODE):
        self.surface_points = surface_points
        self._original_grid = grid
        self._grid = grid
        self.orientations = orientations
        self.unit_values = unit_values
        self.segmentation_function = segmentation_function
        self.stack_relation = stack_relation

    # @ on

    def __repr__(self):
        return pprint.pformat(self.__dict__)

    @classmethod
    def from_interpolation_input_subset(cls, all_interpolation_input: "InterpolationInput",
                                        stack_structure: StacksStructure) -> "InterpolationInput":
        """
        This is the constructor used to extract subsets for each feature/series
        """

        stack_number = stack_structure.stack_number

        sp = SurfacePoints.from_suraface_points_subset(all_interpolation_input.surface_points, stack_structure)
        o = Orientations.from_orientations_subset(all_interpolation_input.orientations, stack_structure)

        cum_number_surfaces_l0 = stack_structure.number_of_surfaces_per_stack[:stack_number].sum()
        cum_number_surfaces_l1 = stack_structure.number_of_surfaces_per_stack[:stack_number + 1].sum() + 1  # * we need to take one unit extra for the basement

        unit_values = all_interpolation_input.unit_values[cum_number_surfaces_l0:cum_number_surfaces_l1]

        grid = all_interpolation_input.grid

        # * (miguel 24) This interpolation input goes on the InterpOutput so we are not computing its gradients
        ii_subset: InterpolationInput = cls(
            surface_points=sp,
            orientations=o,
            grid=grid,
            unit_values=unit_values,
            stack_relation=stack_structure.active_masking_descriptor,
        )

        # ! Setting this on the constructor does not work with data classes.
        ii_subset.fault_values = stack_structure.active_faults_input_data
        ii_subset.all_surface_points = all_interpolation_input.surface_points

        return ii_subset

    @classmethod
    def from_schema(cls, schema: InterpolationInputSchema) -> "InterpolationInput":
        return cls(
            surface_points=SurfacePoints.from_schema(schema.surface_points),
            orientations=Orientations.from_schema(schema.orientations),
            grid=schema.grid,
        )

    @classmethod
    def from_structural_frame(cls, structural_frame: "gempy.StructuralFrame", grid: "gempy.EngineGrid",
                              transform: "gempy.Transfrom") -> "InterpolationInput":
        _legacy_factor = 0

        if LEGACY_COORDS := False:
            _legacy_factor = 0.5

        surface_points: SurfacePoints = SurfacePoints(
            sp_coords=transform.apply(structural_frame.surface_points.xyz) + _legacy_factor,
            nugget_effect_scalar=structural_frame.surface_points.nugget
        )

        orientations: Orientations = Orientations(
            dip_positions=transform.apply(structural_frame.orientations.xyz) + _legacy_factor,
            dip_gradients=transform.transform_gradient(structural_frame.orientations.grads),
            nugget_effect_grad=structural_frame.orientations.nugget
        )

        # region Transforming the grid

        transformed = transform.apply(grid.regular_grid.bounding_box)  # ? isn't this making the regular grid not optional?
        new_extents = np.array([transformed[:, 0].min(), transformed[:, 0].max(),
                                transformed[:, 1].min(), transformed[:, 1].max(),
                                transformed[:, 2].min(), transformed[:, 2].max()])

        # Initialize all variables to None
        octree_grid: Optional[RegularGrid] = None
        regular_grid: Optional[RegularGrid] = None
        custom_values: Optional[GenericGrid] = None
        topography_values: Optional[GenericGrid] = None
        section_values: Optional[GenericGrid] = None
        centered_grid: Optional[CenteredGrid] = None

        if grid.active_grids_bool[0]:
            regular_grid = RegularGrid(
                extent=new_extents,
                regular_grid_shape=grid.regular_grid.resolution
            )

        if grid.active_grids_bool[1] and grid.custom_grid is not None:
            custom_values = GenericGrid(values=transform.apply(grid.custom_grid.values))

        if grid.active_grids_bool[2] and grid.topography is not None:
            topography_values = GenericGrid(values=transform.apply(grid.topography.values))

        if grid.active_grids_bool[3] and grid.sections is not None:
            section_values = GenericGrid(values=transform.apply(grid.sections.values))

        if grid.active_grids_bool[4] and grid.centered_grid is not None:
            centered_grid = CenteredGrid(
                centers=transform.apply(grid.centered_grid.centers),
                radius=transform.scale_points(np.atleast_2d(grid.centered_grid.radius))[0],
                resolution=grid.centered_grid.resolution
            )
        
        octree_grid = RegularGrid(
            extent=new_extents,
            regular_grid_shape=np.array([2, 2, 2])
        )
        
        grid: EngineGrid = EngineGrid(  # * Here we convert the GemPy grid to the
            octree_grid=octree_grid,  # BUG: Adapt the engine to deal with this
            dense_grid=regular_grid,
            topography=topography_values,
            sections=section_values,
            custom_grid=custom_values,
            geophysics_grid=centered_grid
        )

        # endregion

        interpolation_input: InterpolationInput = cls(
            surface_points=surface_points,
            orientations=orientations,
            grid=grid,
            unit_values=structural_frame.elements_ids  # TODO: Here we will need to pass densities etc.
        )

        return interpolation_input

    @property
    def slice_feature(self):
        return self.surface_points.slice_feature

    @property
    def fault_values(self):
        if self._fault_values is None:
            empty_fault_values_on_sp = np.zeros((0, self.surface_points.n_points))
            empty_fault_values_on_grid = np.zeros((0, self.grid.len_all_grids))
            return FaultsData(empty_fault_values_on_grid, empty_fault_values_on_sp)
        return self._fault_values

    @property
    def not_fault_input(self):
        return self._fault_values is None

    @fault_values.setter
    def fault_values(self, value):
        self._fault_values = value

    @property
    def all_surface_points(self):
        if self._all_surface_points is None:
            return self.surface_points  # * This is for backwards compatibility with some tests
        else:
            return self._all_surface_points

    @all_surface_points.setter
    def all_surface_points(self, value):
        self._all_surface_points = value

    @property
    def unit_values(self):
        if self._unit_values is None:
            return np.arange(1000, dtype=np.int16) + 1
        else:
            return self._unit_values.astype(np.int16)

    @unit_values.setter
    def unit_values(self, value):
        self._unit_values = value
